from typing import List, Union
from unittest import TestCase

from autoclass import autoargs, autoprops, getter_override, setter_override, \
    IllegalSetterSignatureException, DuplicateOverrideError, check_var, autoprops_decorate, Boolean, minlens, gts, \
    between, validate, ValidationError, gt


class TestAutoArgs(TestCase):

    def test_autoargs_simple(self):
        """ @autoargs Basic functionality, no customization - all constructor arguments are auto-assigned """

        class A(object):
            @autoargs
            def __init__(self, foo, path, debug=False):
                pass

        # Test :
        # -- create an instance
        a = A('rhubarb', 'pie', debug=True)
        # -- check that the fields exist and have the correct value
        self.assertTrue(a.foo == 'rhubarb')
        self.assertTrue(a.path == 'pie')
        self.assertTrue(a.debug == True)

    def test_autoargs_signature_preserving(self):
        """ @autoargs Advanced: check that the constructor still has the correct signature """

        class A(object):
            @autoargs
            def __init__(self, foo, path, debug=False):
                pass
        try:
            A()
        except TypeError as e:
            self.assertTrue(e.args[0] == "__init__() missing 2 required positional arguments: 'foo' and 'path'")

    def test_autoargs_varargs(self):
        """
        @autoargs Basic functionality, with special case of variable arguments *args.
        -- note that the variable arguments are stored in a single attribute
        """
        class B(object):
            @autoargs
            def __init__(self, foo, path, debug=False, *args):
                pass

        # Test :
        # -- create an instance
        a = B('rhubarb', 'pie', True, 100, 101)
        # -- check that the fields exist and have the correct value
        self.assertTrue(a.foo == 'rhubarb')
        self.assertTrue(a.path == 'pie')
        self.assertTrue(a.debug == True)
        # -- *args is in a single attribute
        self.assertTrue(a.args == (100, 101))

    def test_autoargs_varargs_kwvarargs(self):
        """
        @autoargs Basic functionality, with special case of variable arguments *args and keyword arguments **kw
        -- note that *args are stored in a single attribute while **kw are stored in several attributes
        """
        class C(object):
            @autoargs
            def __init__(self, foo, path, debug=False, *args, **kw):
                pass

        # Test :
        # -- create an instance
        a = C('rhubarb', 'pie', True, 100, 101, verbose=True, bar='bar')
        # -- check that the fields exist and have the correct value
        self.assertTrue(a.foo == 'rhubarb')
        self.assertTrue(a.path == 'pie')
        self.assertTrue(a.debug == True)
        # -- *args is in a single attribute
        self.assertTrue(a.args == (100, 101))
        # -- **kw is dispatched in several attributes
        self.assertTrue(a.verbose == True)
        self.assertTrue(a.bar == 'bar')

    def test_autoargs_noarg(self):
        """ Same than test_autoargs_simple but with empty arguments list in autoargs """

        class O(object):
            @autoargs()
            def __init__(self, foo, path, debug=False):
                pass

        # Test :
        # -- create an instance
        a = O('rhubarb', 'pie', debug=True)
        # -- check that the fields exist and have the correct value
        self.assertTrue(a.foo == 'rhubarb')
        self.assertTrue(a.path == 'pie')
        self.assertTrue(a.debug == True)

    def test_autoargs_include(self):
        """ @autoargs With explicit list of names to include """

        class C(object):
            @autoargs(include=('bar', 'baz', 'verbose'))
            def __init__(self, foo, bar, baz, verbose=False):
                pass

        # Test :
        # -- create an instance
        a = C('rhubarb', 'pie', 1)
        # -- check that the fields exist and have the correct value
        self.assertTrue(a.bar == 'pie')
        self.assertTrue(a.baz == 1)
        self.assertTrue(a.verbose == False)
        # -- check that a non-included field does not exist
        with self.assertRaises(AttributeError):
            print(a.foo)

    def test_autoargs_exclude(self):
        """ @autoargs With explicit list of names to exclude """

        class C(object):
            @autoargs(exclude=('bar', 'baz', 'verbose'))
            def __init__(self, foo, bar, baz, verbose=False):
                pass

        # Test :
        # -- create an instance
        a = C('rhubarb', 'pie', 1)
        # -- check that the fields exist and have the correct value
        self.assertTrue(a.foo == 'rhubarb')
        # -- check that the non-included fields do not exist
        with self.assertRaises(AttributeError):
            print(a.bar)
        with self.assertRaises(AttributeError):
            print(a.baz)
        with self.assertRaises(AttributeError):
            print(a.verbose)

    def test_autoargs_include_exclude(self):
        """ @autoargs assert that you can't use both include/exclude at the same time"""
        with self.assertRaises(ValueError):
            class Dummy(object):
                @autoargs(exclude='', include='')
                def __init__(self, foo, bar, baz, verbose=False):
                    pass

    def test_autoargs_include_exclude_typos(self):
        """ @autoargs Asserts that errors are correctly raised in case of a nonexistent attribute name in
        include/exclude """

        with self.assertRaises(ValueError):
            class Dummy(object):
                @autoargs(exclude='fo')
                def __init__(self, foo, bar, baz, verbose=False):
                    pass

        with self.assertRaises(ValueError):
            class Dummy(object):
                @autoargs(include='fo')
                def __init__(self, foo, bar, baz, verbose=False):
                    pass

    # def test_autoargs_exclude_lock(self):
    #     class C(object):
    #         @autoargs(exclude=('bar', 'baz', 'verbose'), lock_class_fields=True)
    #         def __init__(self, foo, bar, baz, verbose=False):
    #             pass
    #
    #     a = C('rhubarb', 'pie', 1)
    #     self.assertTrue(a.foo == 'rhubarb')
    #     self.assertRaises(AttributeError, getattr, a, 'bar')
    #     self.assertRaises(AttributeError, setattr, a, 'newa',0)


class TestAutoProps(TestCase):

    def test_autoprops_no_contract(self):
        """ Basic @autoprops functionality, no customization - all constructor arguments become properties """

        @autoprops
        class FooConfigA(object):

            @autoargs
            def __init__(self, a: str, b: List[str]):
                pass

        t = FooConfigA('rhubarb', ['pie', 'pie2'])

        # there are no contracts on the generated setters
        t.a=''
        t.b=['r','']
        # check that the generated getters work
        self.assertTrue(t.a == '')
        self.assertTrue(t.b[0] == 'r')

    def test_autoprops_pycontracts(self):
        """
        @autopropsBasic functionality with PyContracts - if a `@contract` annotation exist on the `__init__` method,
        mentioning a contract for a given parameter, the parameter contract will be added on the generated setter method
        """

        from contracts import ContractNotRespected, contract

        @autoprops
        class FooConfigA(object):

            @autoargs
            @contract(a='str[>0]', b='list[>0](str[>0])')
            def __init__(self, a: str, b: List[str]):
                pass

        t = FooConfigA('rhubarb', ['pie', 'pie2'])

        # check that there are contracts on the generated setters
        with self.assertRaises(ContractNotRespected):
            t.a = ''
        with self.assertRaises(ContractNotRespected):
            t.b = ['r','']

        # check that the generated getters work
        t.b=['r']
        self.assertTrue(t.b[0] == 'r')

    def test_autoprops_include(self):
        """ @autoprops With pycontracts and explicit list of attributes to include """

        from contracts import ContractNotRespected, contract

        @autoprops(include='a')
        class FooConfigB(object):

            @autoargs
            @contract(a='str[>0]', b='list[>0](str[>0])')
            def __init__(self, a: str, b: List[str]):
                pass

        t = FooConfigB('rhubarb', ['pie', 'pie2'])

        # check that there is a contract on the generated setter
        with self.assertRaises(ContractNotRespected):
            t.a = ''

        # check that no setter was generated for 'b'
        t.b=[''] # we can because there is no setter, hence no contract
        self.assertTrue(t.b[0] == '')


    def test_autoprops_exclude(self):
        """ @autoprops With pycontracts and explicit list of attributes to exclude """

        from contracts import ContractNotRespected, contract

        @autoprops(exclude='b')
        class FooConfigB(object):

            @autoargs
            @contract(a='str[>0]', b='list[>0](str[>0])')
            def __init__(self, a: str, b: List[str]):
                pass

        t = FooConfigB('rhubarb', ['pie', 'pie2'])

        # check that there is a contract on the generated setter
        with self.assertRaises(ContractNotRespected):
            t.a = ''

        # check that no setter was generated for 'b'
        t.b=[''] # we can because there is no setter, hence no contract
        self.assertTrue(t.b[0] == '')

    def test_autoprops_include_exclude(self):
        """ @autoprops Asserts that include/exclude cant be used at the same time """

        # you can't use both at the same time
        with self.assertRaises(ValueError):
            @autoprops(include='', exclude='')
            class Dummy(object):
                pass

    def test_autoprops_override(self):
        """ @autoprops With Pycontracts. Tests that the user may override generated getter and a setter """

        from contracts import ContractNotRespected, contract

        # check that there is a double-contract warning
        with self.assertWarns(UserWarning):
            @autoprops
            class FooConfigC(object):

                @autoargs
                @contract(a='str[>0]', b='list[>0](str[>0])')
                def __init__(self, a: str, b: List[str]):
                    pass

                @getter_override
                def a(self):
                    # in addition to getting the fields we'd like to print something
                    print('a is being read. Its value is \'' + str(self._a) + '\'')
                    return self._a

                @setter_override(attribute='b')
                @contract(toto='list[>0](str[>0])')
                def another_name(self, toto: List[str]):
                    # in addition to setting the fields we'd like to print something
                    print('Property \'b\' was set to \'' + str(toto) + '\'')
                    self._b = toto

            t = FooConfigC('rhubarb', ['pie', 'pie2'])

            # check that we can still read a's value
            self.assertTrue(t.a == 'rhubarb')

            # check that 'a' still has a contract on its setter
            with self.assertRaises(ContractNotRespected):
                t.a = ''

            # check that 'b' still has a contract on its setter
            with self.assertRaises(ContractNotRespected):
                t.b=[''] # we can not

            # check that 'b' still has a getter generated
            t.b = ['eh', 'oh']
            self.assertTrue(t.b == ['eh', 'oh'])


    def test_autoprops_override_exceptions(self):
        """ @autoprops Asserts that the user can not override a generated method if the overriden method has a wrong
        signature """

        # -- a getter is overriden while the attribute does not exist in constructor
        with self.assertRaises(AttributeError):
            @autoprops
            class FooConfigD(object):
                @autoargs
                def __init__(self, a: str, b: List[str]):
                    pass

                @getter_override
                def c(self):
                    return 'c'

        # -- a setter is overriden with a wrong signature
        with self.assertRaises(IllegalSetterSignatureException):
            @autoprops
            class FooConfigE(object):
                @autoargs
                def __init__(self, a: str, b: List[str]):
                    pass

                @setter_override
                def b(self, toto: List[str], extra_arg:bool = False):
                    self._b = toto

        # -- a getter is overriden twice
        with self.assertRaises(DuplicateOverrideError):
            @autoprops
            class FooConfigF(object):
                @autoargs
                def __init__(self, a: str, b: List[str]):
                    pass

                @getter_override
                def b(self):
                    return self._b

                @getter_override(attribute='b')
                def another(self):
                    return self._b

    def test_autoprops_manual(self):
        """ @autoprops Tests the manual wrapper autoprops() """

        from contracts import ContractNotRespected
        from contracts import contract

        # we don't use @autoprops here
        class FooConfigA(object):
            @autoargs
            @contract(a='str[>0]', b='list[>0](str[>0])')
            def __init__(self, a: str, b: List[str]):
                pass

        # we execute it here
        autoprops_decorate(FooConfigA)

        t = FooConfigA('rhubarb', ['pie', 'pie2'])

        # check that there are contracts on the generated setters
        with self.assertRaises(ContractNotRespected):
            t.a = ''
        with self.assertRaises(ContractNotRespected):
            t.b = ['r', '']

        # check that the generated getters work
        t.b = ['r']
        self.assertTrue(t.b[0] == 'r')

    def test_autoprops_signature_preserving(self):
        """ @autoprops Advanced: checks that the generated constructor still has the correct signature"""

        @autoprops
        class FooConfigD(object):
            @autoargs
            def __init__(self, a: str, b: List[str]):
                pass

        t = FooConfigD('rhubarb', ['pie', 'pie2'])
        try:
            getattr(FooConfigD, 'a').fset(t)
        except TypeError as e:
            self.assertTrue(e.args[0] == "generated_setter_fun() missing 1 required positional argument: 'val'")

        try:
            getattr(FooConfigD, 'a').fget()
        except TypeError as e:
            # yes ; right now it is still a lambda, since using a named method does not seem to work :(
            self.assertTrue(e.args[0] == "<lambda>() missing 1 required positional argument: 'self'")

    def test_autoprops_enforce_validate(self):
        """ Makes sure that autoprops works with enforce AND validate """

        # from autoclass import autoargs, autoprops, Boolean
        import enforce as en
        from enforce import runtime_validation
        from numbers import Real, Integral
        from typing import Optional

        en.config(dict(mode='covariant'))  # allow subclasses when validating types

        @runtime_validation
        @autoprops
        class HouseConfiguration(object):

            @autoargs
            @validate(name=minlens(0),
                      surface=gts(0),
                      nb_floors=between(1, 100, open_right=True))
            def __init__(self,
                         name: str,
                         surface: Real,
                         nb_floors: Optional[Integral] = 1,
                         with_windows: Boolean = False):
                pass

            # -- overriden setter for surface for custom validation or other things
            @setter_override
            def surface(self, surface):
                print('Set surface to {}'.format(surface))
                self.toto = 'done'
                self._surface = surface

        t = HouseConfiguration('test', 12, 2)

        # 'Optional' works
        t.nb_floors = None

        # Custom print works
        t.surface = 0.1
        assert t.toto == 'done'

        # Type validation works
        from enforce.exceptions import RuntimeTypeError
        with self.assertRaises(RuntimeTypeError):
            t.nb_floors = 2.2

        # Custom validation works
        with self.assertRaises(ValidationError):
            t.surface = 0

class TestReadMe(TestCase):
    """
    just to test that what is written in the readme actually works :)
    """

    def test_readme_old_way(self):
        """ Makes sure that the code in the documentation page is correct for the 'old way' of writing classes """

        from autoclass import check_var, Boolean
        from numbers import Real, Integral
        from typing import Optional, Union

        class HouseConfiguration(object):
            def __init__(self,
                         name: str,
                         surface: Real,
                         nb_floors: Optional[Integral] = 1,
                         with_windows: Boolean = False):
                self.name = name
                self.surface = surface
                self.nb_floors = nb_floors
                self.with_windows = with_windows

            # --name
            @property
            def name(self):
                return self._name

            @name.setter
            def name(self, name: str):
                check_var(name, var_name='name', var_types=str)
                self._name = name

            # --surface
            @property
            def surface(self) -> Real:
                return self._surface

            @surface.setter
            def surface(self, surface: Real):
                check_var(surface, var_name='surface', var_types=Real, min_value=0, min_strict=True)
                self._surface = surface

            # --nb_floors
            @property
            def nb_floors(self) -> Optional[Integral]:
                return self._nb_floors

            @nb_floors.setter
            def nb_floors(self, nb_floors: Optional[Integral]):
                check_var(nb_floors, var_name='nb_floors', var_types=Integral, enforce_not_none=False)
                self._surface = nb_floors  # !**

            # --with_windows
            @property
            def with_windows(self) -> Boolean:
                return self._with_windows

            @with_windows.setter
            def with_windows(self, with_windows: Boolean):
                check_var(with_windows, var_name='with_windows', var_types=Boolean)
                self._with_windows = with_windows

        HouseConfiguration('test', 0.1)

    def test_readme_pycontracts(self):
        """ Makes sure that the code in the documentation page is correct for the PyContracts example """

        from contracts import contract, ContractNotRespected
        from numbers import Real, Integral
        from typing import Optional

        @autoprops
        class HouseConfiguration(object):
            @autoargs
            @contract(name='str[>0]',
                      surface='(int|float),>=0',
                      nb_floors='None|int',
                      with_windows='bool')
            def __init__(self,
                         name: str,
                         surface: Real,
                         nb_floors: Optional[Integral] = 1,
                         with_windows: Boolean = False):
                pass

            # -- overriden setter for surface - no need to repeat the @contract
            @setter_override
            def surface(self, surface: Real):
                assert surface > 0
                self._surface = surface

        t = HouseConfiguration('test', 0.1)
        t.nb_floors = None
        with self.assertRaises(ContractNotRespected):
            t.nb_floors = 2.2
        with self.assertRaises(ContractNotRespected):
            t.surface = -1

    def test_readme_enforce(self):
        """ Makes sure that the code in the documentation page is correct for the enforce example """

        # from autoclass import autoargs, autoprops, Boolean
        import enforce as en
        from enforce import runtime_validation
        from numbers import Real, Integral
        from typing import Optional

        en.config(dict(mode='covariant'))  # allow subclasses when validating types

        @runtime_validation
        @autoprops
        class HouseConfiguration(object):
            @autoargs
            def __init__(self,
                         name: str,
                         surface: Real,
                         nb_floors: Optional[Integral] = 1,
                         with_windows: Boolean = False):
                pass

            # -- overriden setter for surface for custom validation
            @setter_override
            def surface(self, surface):
                assert surface > 0
                self._surface = surface

        t = HouseConfiguration('test', 12, 2)

        # 'Optional' works
        t.nb_floors = None

        # Type validation works
        from enforce.exceptions import RuntimeTypeError
        with self.assertRaises(RuntimeTypeError):
            t.nb_floors = 2.2

        # Custom validation works
        with self.assertRaises(AssertionError):
            t.surface = 0

    def test_readme_enforce_validate(self):
        """ Makes sure that the code in the documentation page is correct for the enforce + validate example """

        # from autoclass import autoargs, autoprops, Boolean
        import enforce as en
        from enforce import runtime_validation
        from numbers import Real, Integral
        from typing import Optional

        en.config(dict(mode='covariant'))  # allow subclasses when validating types

        @runtime_validation
        @autoprops
        class HouseConfiguration(object):
            @autoargs
            @validate(name=minlens(0),
                      surface=gt(0))
            def __init__(self,
                         name: str,
                         surface: Real,
                         nb_floors: Optional[Integral] = 1,
                         with_windows: Boolean = False):
                pass

            # -- overriden setter for surface for custom validation or other things
            @setter_override
            def surface(self, surface):
                print('Set surface to {}'.format(surface))
                self._surface = surface

        t = HouseConfiguration('test', 12, 2)

        # 'Optional' works
        t.nb_floors = None

        # Type validation works
        from enforce.exceptions import RuntimeTypeError
        with self.assertRaises(RuntimeTypeError):
            t.nb_floors = 2.2

        # Value validation works
        with self.assertRaises(ValidationError):
            t.surface = -1

        # Value validation works in constructor
        with self.assertRaises(ValidationError):
            HouseConfiguration('', 12, 2)

    def test_readme_usage_autoprops_validate(self):
        @autoprops
        class FooConfigA(object):
            @autoargs
            @validate(a=minlens(0))
            def __init__(self, a: str):
                pass

        t = FooConfigA('rhubarb')

        # check that the generated getters work
        t.a = 'r'
        assert t.a == 'r'

        # check that there are validators on the generated setters
        with self.assertRaises(ValidationError):
            t.a = ''  # raises ValidationError



from unittest import TestCase

from autoclass import validate, ValidationError, is_even, gt, not_none, not_, is_mod, or_, xor_, is_subset, and_, \
    is_superset, is_in


class TestValidate(TestCase):

    def test_validate_simple(self):
        """ Simple test of the @validate annotation, with built-in validators is_even and gt(1) """

        @validate(a=[not_none, is_even, gt(1)],
                  b=is_even)
        def myfunc(a, b):
            print('hello')

        # -- check that the validation works
        myfunc(84, 82)
        with self.assertRaises(ValidationError):
            # a is None
            myfunc(None,0)
        with self.assertRaises(ValidationError):
            # a is not even
            myfunc(1,0)
        with self.assertRaises(ValidationError):
            # b is not even
            myfunc(2,1)
        with self.assertRaises(ValidationError):
            # a is not >= 1
            myfunc(0,0)

    def test_validate_wrong_notnone(self):
        with self.assertRaises(ValueError):
            @validate(a=[is_even, not_none, gt(1)],
                      b=is_even)
            def myfunc(a, b):
                print('hello')

    def test_validate_empty(self):
        with self.assertRaises(ValueError):
            @validate(a=[],
                      b=is_even)
            def myfunc(a, b):
                print('hello')

    def test_validate_custom(self):
        """ Simple test of the @validate annotation, with custom validators of several styles"""

        def is_mod_3(x):
            """ A simple validator with no parameters"""
            return x % 3 == 0

        def is_mod(ref):
            """ A validator generator, with parameters """
            def is_mod_ref(x):
                return x % ref == 0
            return is_mod_ref

        def gt_ex1(x):
            """ A validator raising a custom exception in case of failure """
            if x >= 1:
                return True
            else:
                raise ValidationError('gt_ex1: x >= 1 does not hold for x=' + str(x))

        def gt_assert2(x):
            """ (not recommended) A validator relying on assert and therefore only valid in 'debug' mode """
            assert x >= 2

        @validate(a=[gt_ex1, gt_assert2, is_mod_3],
                  b=is_mod(5))
        def myfunc(a, b):
            print('hello')

        # -- check that the validation works
        myfunc(21, 15)
        with self.assertRaises(ValidationError):
            myfunc(4,21)  # ValidationError: a is not a multiple of 3
        with self.assertRaises(ValidationError):
            myfunc(15,1)  # ValidationError: b is not a multiple of 5
        with self.assertRaises(AssertionError):
            myfunc(1,0)  # AssertionError: a is not >= 2
        with self.assertRaises(ValidationError):
            myfunc(0,0)  # ValidationError: a is not >= 1

    def test_validate_enforce(self):
        """ Tests that a None will be catched by enforce: no need for not_none validator """
        from enforce import runtime_validation, config
        from enforce.exceptions import RuntimeTypeError
        from numbers import Integral

        # we're not supposed to do that but if your python environment is a bit clunky, that might help
        # config(dict(mode='covariant'))

        @runtime_validation
        @validate(a=[is_even, gt(1)], b=is_even)
        def myfunc(a: Integral, b):
            print('hello')

        # -- check that the validation works
        myfunc(84, None)  # OK because b has no type annotation nor not_none validator
        with self.assertRaises(RuntimeTypeError):
            myfunc(None, 0)  # RuntimeTypeError: a is None

    def test_validate_name_error(self):
        """ Checks that wrong attribute names cant be provided to @validate"""
        with self.assertRaises(ValueError):
            @validate(ab=[])
            def myfunc(a, b):
                print('hello')

    def test_validate_none_is_allowed(self):
        @validate(a=is_even)
        def myfunc(a, b):
            print('hello')

        # -- check that the validation works
        myfunc(84, 82)
        myfunc(None, 0)

    def test_validate_not(self):
        """ Test for the not_ validator """

        def gtcustom(x):
            assert x < 10

        @validate(a=not_(is_even), b=not_([is_even, is_mod(3)]), c=not_(gtcustom, catch_all=True),
                  d=not_(gtcustom))
        def myfunc(a, b, c, d):
            print('hello')

        # -- check that the validation works
        myfunc(11, 11, 11, None)

        with self.assertRaises(ValidationError):
            myfunc(84, 82, None, None)  # ValidationError: a is even

        with self.assertRaises(ValidationError):
            myfunc(84, 3, None, None)  # ValidationError: b is odd (ok) but it is a multiple of 3 (nok)

        with self.assertRaises(ValidationError):
            myfunc(11, 11, 9, 11)  # c is not valid but the not_ operator catches the exception and wraps it

        with self.assertRaises(ValidationError):
            myfunc(11, 11, 11, 9)  # d is not valid

        with self.assertRaises(AssertionError):
            myfunc(11, 11, 11, 11)  # d is valid but the not_ operator does not catch the exception so we get the error


    def test_validate_or(self):
        """ Test for the or_ validator, also in combination with not_"""

        # empty list error
        with self.assertRaises(ValueError):
            @validate(a=or_([]))
            def myfunc(a, b):
                print('hello')

        # single element simplification
        @validate(a=or_([is_even]))
        def myfunc(a):
            print('hello')

        myfunc(4)
        with self.assertRaises(ValidationError):
            myfunc(7)

        # lists
        @validate(a=or_([is_even, is_mod(3)]), b=not_(or_([is_even, is_mod(3)])))
        def myfunc(a, b):
            print('hello')

        # -- check that the validation works
        myfunc(9, None)  # a is not even but is a multiple of 3 > ok
        myfunc(4, None)  # a is even but is not a multiple of 3 > ok
        myfunc(6, 7)     # b is not even AND not a multiple of 3 > ok

        with self.assertRaises(ValidationError):
            myfunc(7, None)  # ValidationError: a is odd and not multiple of 3

        with self.assertRaises(ValidationError):
            myfunc(None, 3)  # ValidationError: b is odd but it is a multiple of 3

    def test_validate_xor(self):
        """ Test for the xor_ validator """

        # empty list error
        with self.assertRaises(ValueError):
            @validate(a=xor_([]))
            def myfunc(a, b):
                print('hello')

        # single element simplification
        @validate(a=xor_([is_even]))
        def myfunc(a):
            print('hello')

        myfunc(4)
        with self.assertRaises(ValidationError):
            myfunc(7)

        # lists
        @validate(a=xor_([is_even, is_mod(3)]))
        def myfunc(a):
            print('hello')

        # -- check that the validation works
        myfunc(9)  # a is not even but is a multiple of 3 > ok
        myfunc(4)  # a is even but is not a multiple of 3 > ok

        with self.assertRaises(ValidationError):
            myfunc(6)  # ValidationError: a is both even and a multiple of 3

        with self.assertRaises(ValidationError):
            myfunc(7)  # ValidationError: a is both even and a multiple of 3

    def test_validate_and(self):
        """ Simple test of the @validate annotation, with built-in validators is_even and gt(1) """

        # empty list error
        with self.assertRaises(ValueError):
            @validate(a=and_([]))
            def myfunc(a, b):
                print('hello')

        # single element simplification
        @validate(a=and_([is_even]))
        def myfunc(a):
            print('hello')

        myfunc(4)
        with self.assertRaises(ValidationError):
            myfunc(7)

        # lists
        @validate(a=[not_none, and_([is_even, gt(1)])])
        def myfunc(a):
            print('hello')

        # -- check that the validation works
        myfunc(84)
        with self.assertRaises(ValidationError):
            # a is None
            myfunc(None)
        with self.assertRaises(ValidationError):
            # a is not even
            myfunc(1)
        with self.assertRaises(ValidationError):
            # a is not >= 1
            myfunc(0)

    def test_validate_is_in(self):
        """ Test for the subset and superset validators """

        @validate(a=is_in({'+', '-'}))
        def myfunc(a):
            print('hello')

        # -- check that the validation works
        myfunc('+')
        with self.assertRaises(ValidationError):
            myfunc('*')

    def test_validate_subset_superset(self):
        """ Test for the subset and superset validators """

        @validate(a=is_subset({'+', '-'}), b=is_superset({'+', '-'}),
                  c=[is_subset({'+', '-'}), is_superset({'+', '-'})])
        def myfunc(a, b, c):
            print('hello')

        # -- check that the validation works
        myfunc({'+'},{'+', '-', '*'}, {'+', '-'})

        with self.assertRaises(ValidationError):
            myfunc({'+', '-', '*'}, None, None)

        with self.assertRaises(ValidationError):
            myfunc(None, {'+'}, None)

        with self.assertRaises(ValidationError):
            myfunc(None, None, {'+', '-', '*'})

        with self.assertRaises(ValidationError):
            myfunc(None, None, {'+'})



from inspect import getfullargspec
from numbers import Integral
from typing import Callable, Dict, Any, Set

from decorator import decorate

from autoclass.utils import _create_function_decorator__robust_to_args, apply_on_func_args


def validate(**validators: Dict[str, Callable[[Any], bool]]):
    """
    Defines a decorator with parameters, that will execute the provided input validators PRIOR to executing the 
    function. Specific entry 'returns' may contain validators executed AFTER executing the function.
    
    ```
    def is_even(x):
        return x % 2 == 0
    
    def gt(a):
        def gt(x):
            return x >= a
        return gt
    
    @validate(a=[is_even, gt(1)], b=is_even)
    def myfunc(a, b):
        print('hello')
    ```
    
    will generate the equivalent of :
    
    ```
    def myfunc(a, b):
        gt1 = gt(1)
        if is_even(a) and gt1(a) and is_even(b):
            print('hello')
        else:
            raise ValidationError(...)
    ```
    
    :param validators: 
    :return: 
    """
    return _create_function_decorator__robust_to_args(validate_decorate, **validators)


def validate_decorate(func: Callable, **validators: Dict[str, Callable[[Any], bool]]) -> Callable:
    """
    Defines a decorator with parameters, that will execute the provided input validators PRIOR to executing the 
    function. Specific entry 'returns' may contain validators executed AFTER executing the function.
    
    :param func: 
    :param include: 
    :param exclude: 
    :return: 
    """
    # (1) retrieve function signature
    # attrs, varargs, varkw, defaults = getargspec(func)
    signature_attrs, signature_varargs, signature_varkw, signature_defaults, signature_kwonlyargs, \
    signature_kwonlydefaults, signature_annotations = getfullargspec(func)
    # TODO better use signature(func) ? but that would be less compliant with python 2

    # (2) check that provided validators dont contain names that are incorrect
    if validators is not None:
        incorrect = set(validators.keys()) - set(signature_attrs)
        if len(incorrect) > 0:
            raise ValueError('@validate definition exception: validators are defined for \'' + str(incorrect) + '\' '
                             'that is/are not part of signature for ' + str(func))
    # for att_name, att_validators in validators.items():
    #     i = att_validators.index(not_none)
    #     if i > 0:
    #         raise ValueError('not_none is a special validator that can only be provided at the beginning of the'
    #                          ' validators list')

    # replace validators lists with explicit and_ if needed
    for att_name, att_validators in validators.items():
        att_validators = _assert_list_and_protect_not_none(att_validators, allow_not_none=True)
        if att_validators[0] != not_none:
            # att_validators is a list not containing not_none: wrap in an 'and_' and wrap with a none-ignoring checker
            validators[att_name] = _not_none_checker(and_(att_validators), ignore_none_silently=True)
        else:
            # first element of att_validators is not_none
            if len(att_validators) > 1:
                # remove not_none from the list, wrap witha and_, and wrap with the corresponding_not_none_checker
                validators[att_name] = _not_none_checker(and_(att_validators[1:]), ignore_none_silently=False)
            else:
                # one element in the list only (not_none): include it directly
                validators[att_name] = att_validators[0]

    # (3) create a wrapper around the function to add validation
    # -- old:
    # @functools.wraps(func) -> to make the wrapper function look like the wrapped function
    # def wrapper(self, *args, **kwargs):
    # -- new:
    # we now use 'decorate' at the end of this code to have a wrapper that has the same signature, see below
    def wrapper(func, *args, **kwargs):
        # apply _validate on all received arguments
        apply_on_func_args(func, args, kwargs, signature_attrs, signature_defaults, signature_varargs, signature_varkw,
                           _validate, validators)

        # finally execute the method
        return func(*args, **kwargs)

    a = decorate(func, wrapper)
    # save the validators somewhere for reference
    a.__validators__ = validators
    return a


def _not_none_checker(validator, ignore_none_silently: bool = True):
    """
    Generates a checker handling None values. When a None value is received, it is not passed to the validator.
    Instead this operator will either drop silently (ignore_none_silently = True) or return a False
    (ignore_none_silently=False)

    :param validator:
    :param ignore_none_silently:
    :return:
    """
    if ignore_none_silently:
        def drop_none_silently(x):
            if not_none(x):
                return validator(x)
            else:
                # value is None : skip validation (users should explicitly include 'not_none' as the first validator to
                # change this behaviour)
                return True
        return drop_none_silently
    else:
        def check_not_none(x):
            if not_none(x):
                return validator(x)
            else:
                return False
        return check_not_none


def _validate(value_to_validate, validator_func, func, att_name):
    """
    Subroutine that actually executes validation

    :param value_to_validate: the value to validate
    :param validator_func: the validator function that will be applied on value_to_validate
    :param func: the method for which this validation is performed. This is used just for errors
    :param att_name: the name of the attribute that is being validated
    :return:
    """
    # new: validator_func should always be a single element here
    res = validator_func(value_to_validate)
    if res not in {None, True}:
        raise ValidationError.create(func, att_name, validator_func, value_to_validate)


class ValidationError(Exception):
    """
    Exception raised whenever validation fails. It may be directly triggered by Validators, or it is raised if 
    validator returns false
    """

    def __init__(self, contents):
        """
        We actually can't put more than 1 argument in the constructor, it creates a bug in Nose tests
        https://github.com/nose-devs/nose/issues/725
        
        Please use ValidationError.create() instead

        :param contents:
        """
        super(ValidationError, self).__init__(contents)

    @staticmethod
    def create(func, att_name, validator_func, item, extra_msg: str = None):
        """
        
        :param func:
        :param att_name:
        :param validator_func: 
        :param item: 
        :param extra_msg
        :return: 
        """
        return ValidationError('Error validating input ' + str(att_name) + '=\'' + str(item) + '\' for function \''
                               + str(func) + '\' with validator ' + (validator_func.__name__ or str(validator_func))
                               + (('.\n  ' + extra_msg) if extra_msg else ''))


def get_names(validators):
    return ', '.join([val.__name__ for val in validators])


# ----------- some convenient validators ...
def not_none(x: Any):
    """ 'Is not None' validator """
    return x is not None


def and_(validators):
    """
    An 'and' validator: returns True if all of the provided validators are happy with the input. This method will
    raise a ValidationException on the first False received or Exception caught

    :param validators:
    :return:
    """

    validators = _assert_list_and_protect_not_none(validators)

    if len(validators) == 1:
        return validators[0]  # simplification for single validator case
    else:
        def and_v_(x):
            for validator in validators:
                res = validator(x)
                if res not in {None, True}:
                    # one validator was unhappy > raise
                    raise ValidationError('and(' + get_names(validators) + '): Validator ' + str(validator)
                                          + ' failed validation for input ' + str(x))
            return True

        return and_v_


def or_(validators):
    """
    An 'or' validator: returns True if at least one of the provided validators is happy with the input. All exceptions
    will be silently caught. In case of failure, a global ValidationException will be raised, together with the first
    exception message if any.

    :param validators:
    :return:
    """

    validators = _assert_list_and_protect_not_none(validators)

    if len(validators) == 1:
        return validators[0]  # simplification for single validator case
    else:
        def or_v_(x):
            err = None
            for validator in validators:
                try:
                    res = validator(x)
                    if res in {None, True}:
                        # we can return : one validator was happy
                        return True
                except Exception as e:
                    if err is None:
                        err = e  # remember the first exception

            # no validator accepted: raise
            msg = 'or(' + get_names(validators) + '): All validators failed validation for input \'' + str(x) + '\'. '
            if err is not None:
                msg += 'First exception caught was: \'' + str(err) + '\''
            raise ValidationError(msg)

        return or_v_


def xor_(validators):
    """
    A 'xor' validator: returns True if exactly one of the provided validators is happy with the input. All
    exceptions will be silently caught. In case of failure, a global ValidationException will be raised, together with
    the first exception message if any.

    :param validators:
    :return:
    """

    validators = _assert_list_and_protect_not_none(validators)

    if len(validators) == 1:
        return validators[0]  # simplification for single validator case
    else:
        def xor_v_(x):
            ok_validator = None
            sec_validator = None
            err = None
            for validator in validators:
                try:
                    res = validator(x)
                    if res in {None, True}:
                        if ok_validator is not None:
                            sec_validator = validator
                        else:
                            # we found the first one happy
                            ok_validator = validator
                except Exception as e:
                    if err is None:
                        err = e  # remember the first exception

            # return if were happy or not
            if ok_validator is not None:
                if sec_validator is None:
                    # one unique validator happy: success
                    return True
                else:
                    # second validator happy : fail, too many validators happy
                    raise ValidationError('xor(' + get_names(validators) + ') : Too many validators succeeded : '
                                          + str(ok_validator) + ' + ' + str(sec_validator))
            else:
                # no validator happy
                msg = 'xor(' + get_names(validators) + '): All validators failed validation for input \'' + str(x) + '\'. '
                if err is not None:
                    msg += 'First exception caught was: \'' + str(err) + '\''
                raise ValidationError(msg)

        return xor_v_


def _assert_list_and_protect_not_none(validators, allow_not_none: bool = False):
    """
     * if validators is an empty list, throw error
     * If validators is a singleton, turns it into a list.
     * If validators contains not_none and allow_not_none is set to True, asserts that not_none is first in the list
     * If validators contains not_none and allow_not_none is set to False, asserts that not_none is not present at all
     in the list

    :param validators:
    :param allow_not_none:
    :return:
    """
    i = -1
    try:
        i = validators.index(not_none)
    except ValueError:
        # not_none not found in validators list : ok
        pass
    except AttributeError:
        # validators is not a list (no attribute 'index'): turn it into a list
        validators = [validators]

    # not_none ?
    if i > 0 or (i == 0 and not allow_not_none):
        raise ValueError('not_none is a special validator that can only be provided at the beginning of the'
                         ' global validators list')
    # empty list ?
    if len(validators) == 0:
        raise ValueError('provided validators list is empty')

    return validators


def not_(validator, catch_all: bool = False):
    """
    Generates the inverse of the provided validator: when the validator returns False or raises a `ValidationError`,
    this validator returns True. Otherwise it returns False. If the `catch_all` parameter is set to `True`, any
    exception will be caught instead of just `ValidationError`s.

    Note that you may provide a list of validators to `not_()`. It will be interpreted as `not_(and_(<valiators_list>))`

    :param validator:
    :param catch_all: an optional boolean flag. By default, only ValidationError are silently catched and turned into
    a 'ok' result. Turning this flag to True will assume that all exceptions should be catched and turned to a
    'ok' result
    :return:
    """

    # in case this is a validator list, create a 'and_' around it (otherwise this returns the validator)
    validator = and_(validator)

    def not_v_(x):
        if catch_all:
            try:
                res = validator(x)
                if res not in {None, True}:  # inverse the result
                    return True
            except:
                return True  # caught exception: return True

            # if we're here that's a failure
            raise ValidationError('not(' + str(validator) + '): Validator validated input \'' + str(x) + '\' with success, '
                                  'therefore the not() is a failure')
        else:
            try:
                res = validator(x)
                if res not in {None, True}:  # inverse the result
                    return True
            except ValidationError:
                return True  # caught exception: return True

            # if we're here that's a failure
            raise ValidationError('not(' + str(validator) + '): Validator validated input \'' + str(x) + '\' with success, '
                                  'therefore the not() is a failure')
    return not_v_


# ------------- orderables ----------------
def gt(min_value: Any, strict: bool = False):
    """
    'Greater than' validator generator.
    Returns a validator to check that x >= min_value (strict=False, default) or x > min_value (strict=True)

    :param min_value: minimum value for x
    :param strict: Boolean flag to switch between x >= min_value (strict=False) and x > min_value (strict=True)
    :return:
    """
    if strict:
        def gt(x):
            if x > min_value:
                return True
            else:
                raise ValidationError('gt: x > ' + str(min_value) + ' does not hold for x=' + str(x))
    else:
        def gt(x):
            if x >= min_value:
                return True
            else:
                raise ValidationError('gt: x >= ' + str(min_value) + ' does not hold for x=' + str(x))
    return gt


def gts(min_value_strict: Any):
    """ Alias for 'greater than' validator generator in strict mode """
    return gt(min_value_strict, True)


def lt(max_value: Any, strict: bool = False):
    """
    'Lesser than' validator generator.
    Returns a validator to check that x <= max_value (strict=False, default) or x < max_value (strict=True)

    :param max_value: maximum value for x
    :param strict: Boolean flag to switch between x <= max_value (strict=False) and x < max_value (strict=True)
    :return:
    """
    if strict:
        def lt(x):
            if x < max_value:
                return True
            else:
                raise ValidationError('lt: x < ' + str(max_value) + ' does not hold for x=' + str(x))
    else:
        def lt(x):
            if x <= max_value:
                return True
            else:
                raise ValidationError('lt: x <= ' + str(max_value) + ' does not hold for x=' + str(x))
    return lt


def lts(max_value_strict: Any):
    """ Alias for 'lesser than' validator generator in strict mode """
    return gt(max_value_strict, True)


def between(min_val: Any, max_val: Any, open_left: bool = False, open_right: bool = False):
    """
    'Is between' validator generator.
    Returns a validator to check that min_val <= x <= max_val (default). open_right and open_left flags allow to
    transform each side into strict mode. For example setting open_left=True will enforce min_val < x <= max_val

    :param min_val: minimum value for x
    :param max_val: maximum value for x
    :param open_left: Boolean flag to turn the left inequality to strict mode
    :param open_right: Boolean flag to turn the right inequality to strict mode
    :return:
    """
    if open_left and open_right:
        def between(x):
            if (min_val < x) and (x < max_val):
                return True
            else:
                raise ValidationError('between: ' + str(min_val) + ' < x < ' + str(max_val) + ' does not hold for x='
                                      + str(x))
    elif open_left:
        def between(x):
            if (min_val < x) and (x <= max_val):
                return True
            else:
                raise ValidationError('between: ' + str(min_val) + ' < x <= ' + str(max_val) + ' does not hold for x='
                                      + str(x))
    elif open_right:
        def between(x):
            if (min_val <= x) and (x < max_val):
                return True
            else:
                raise ValidationError('between: ' + str(min_val) + ' <= x < ' + str(max_val) + ' does not hold for x='
                                      + str(x))
    else:
        def between(x):
            if (min_val <= x) and (x <= max_val):
                return True
            else:
                raise ValidationError('between: ' + str(min_val) + ' <= x <= ' + str(max_val) + ' does not hold for x='
                                      + str(x))
    return between


# ------------- integers ------------------
def is_even(x: Integral):
    """ Validates that x is even (`x % 2 == 0`) """
    return x % 2 == 0


def is_odd(x: Integral):
    """ Validates that x is odd (`x % 2 != 0`) """
    return x % 2 != 0


def is_mod(ref):
    """ Validates that x is a multiple of the reference (`x % ref == 0`) """
    def is_mod(x):
        if x % ref == 0:
            return True
        else:
            raise ValidationError('is_mod: x % ' + str(ref) + ' == 0 does not hold for x=' + str(x))
    return is_mod


# ------------- collections ----------------
def minlen(min_length: Integral, strict: bool = False):
    """
    'Minimum length' validator generator.
    Returns a validator to check that len(x) >= min_length (strict=False, default) or len(x) > min_length (strict=True)

    :param min_length: minimum length for x
    :param strict: Boolean flag to switch between len(x) >= min_length (strict=False) and len(x) > min_length
    (strict=True)
    :return:
    """
    if strict:
        def minlen(x):
            if len(x) > min_length:
                return True
            else:
                raise ValidationError('minlen: len(x) > ' + str(min_length) + ' does not hold for x=' + str(x))
    else:
        def minlen(x):
            if len(x) >= min_length:
                return True
            else:
                raise ValidationError('minlen: len(x) >= ' + str(min_length) + ' does not hold for x=' + str(x))
    return minlen


def minlens(min_length_strict: Integral):
    """ Alias for 'Minimum length' validator generator in strict mode """
    return minlen(min_length_strict, True)


def maxlen(max_length: Integral, strict: bool = False):
    """
    'Maximum length' validator generator.
    Returns a validator to check that len(x) <= max_length (strict=False, default) or len(x) < max_length (strict=True)

    :param max_length: maximum length for x
    :param strict: Boolean flag to switch between len(x) <= max_length (strict=False) and len(x) < max_length
    (strict=True)
    :return:
    """
    if strict:
        def maxlen(x):
            if len(x) < max_length:
                return True
            else:
                raise ValidationError('maxlen: len(x) < ' + str(max_length) + ' does not hold for x=' + str(x))
    else:
        def maxlen(x):
            if len(x) <= max_length:
                return True
            else:
                raise ValidationError('maxlen: len(x) <= ' + str(max_length) + ' does not hold for x=' + str(x))
    return maxlen


def maxlens(max_length_strict: Integral):
    """ Alias for 'Maximum length' validator generator in strict mode """
    return maxlen(max_length_strict, True)


def is_in(allowed_values: Set):
    """
    'Values in' validator generator.
    Returns a validator to check that x is in the provided set of allowed values

    :param allowed_values: a set of allowed values
    :return:
    """
    def valin(x):
        if x in allowed_values:
            return True
        else:
            raise ValidationError('is_in: x in ' + str(allowed_values) + ' does not hold for x=' + str(x))
    return valin


def is_subset(reference_set: Set):
    """
    'Is subset' validator generator.
    Returns a validator to check that x is a subset of reference_set

    :param reference_set: the reference set
    :return:
    """
    def is_subset(x):
        missing = x - reference_set
        if len(missing) == 0:
            return True
        else:
            raise ValidationError('is_subset: len(x - reference_set) == 0 does not hold for x=' + str(x)
                                  + ' and reference_set=' + str(reference_set) + '. x contains unsupported '
                                  'elements ' + str(missing))
    return is_subset


def is_superset(reference_set: Set):
    """
    'Is superset' validator generator.
    Returns a validator to check that x is a superset of reference_set

    :param reference_set: the reference set
    :return:
    """
    def is_superset(x):
        missing = reference_set - x
        if len(missing) == 0:
            return True
        else:
            raise ValidationError('is_superset: len(reference_set - x) == 0 does not hold for x=' + str(x)
                                  + ' and reference_set=' + str(reference_set) + '. x does not contain required '
                                  'elements ' + str(missing))
    return is_superset



from autoclass.autoargs import *
from autoclass.autoprops import *
from autoclass.var_checker import *
from autoclass.validate import *
from autoclass.booleans import *

__all__ = ['autoargs', 'autoprops', 'validate', 'var_checker', 'booleans']


from abc import abstractmethod, ABCMeta


class Boolean(metaclass=ABCMeta):
    """
    An abstract base class for booleans, similar to what is available in numbers
    see https://docs.python.org/3.5/library/numbers.html
    """
    __slots__ = ()

    @abstractmethod
    def __bool__(self):
        """Return a builtin bool instance. Called for bool(self)."""

    @abstractmethod
    def __and__(self, other):
        """self & other"""

    @abstractmethod
    def __rand__(self, other):
        """other & self"""

    @abstractmethod
    def __xor__(self, other):
        """self ^ other"""

    @abstractmethod
    def __rxor__(self, other):
        """other ^ self"""

    @abstractmethod
    def __or__(self, other):
        """self | other"""

    @abstractmethod
    def __ror__(self, other):
        """other | self"""

    @abstractmethod
    def __invert__(self):
        """~self"""

# register bool and numpy bool_ as virtual subclasses
# so that issubclass(bool, Boolean) = issubclass(np.bool_, Boolean) = True
Boolean.register(bool)

try:
    import numpy as np
    Boolean.register(np.bool_)
except ImportError:
    # silently escape
    pass
